# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module handles photometry units that are not in `astropy.units`."""

# ASTROPY
from astropy import constants as const
from astropy import units as u

# LOCAL
from synphot import exceptions

__all__ = ['H', 'C', 'HC', 'SR_PER_ARCSEC2', 'AREA', 'THROUGHPUT', 'PHOTLAM',
           'PHOTNU', 'FLAM', 'FNU', 'OBMAG', 'VEGAMAG',
           'spectral_density_vega', 'spectral_density_count', 'convert_flux',
           'validate_unit', 'validate_wave_unit', 'validate_quantity']

# ----------------- #
# General constants #
# ----------------- #

H = const.h.cgs  # Planck's constant in erg * sec
C = const.c.to('AA/s')  # Speed of light in Angstrom/sec
HC = H * C
SR_PER_ARCSEC2 = u.rad.to(u.arcsec) ** -2  # steradian per arcsec^2

# ------------- #
# synphot units #
# ------------- #

# Default unit of area covered by flux
AREA = u.cm * u.cm

# synphot unitless unit (using def_unit mess up arithmetic result unit string)
THROUGHPUT = u.dimensionless_unscaled

# synphot flux units
PHOTLAM = u.def_unit(
    'photlam', u.photon / (u.cm**2 * u.s * u.AA),
    format={'generic': 'PHOTLAM', 'console': 'PHOTLAM'})
PHOTNU = u.def_unit(
    'photnu', u.photon / (u.cm**2 * u.s * u.Hz),
    format={'generic': 'PHOTNU', 'console': 'PHOTNU'})
FLAM = u.def_unit(
    'flam', u.erg / (u.cm**2 * u.s * u.AA),
    format={'generic': 'FLAM', 'console': 'FLAM'})
FNU = u.def_unit(
    'fnu', u.erg / (u.cm**2 * u.s * u.Hz),
    format={'generic': 'FNU', 'console': 'FNU'})
_u_ob = u.def_unit('OB')
OBMAG = u.mag(_u_ob)
_u_vega = u.def_unit('VEGA')
VEGAMAG = u.mag(_u_vega)

# Register with astropy units
u.add_enabled_units([PHOTLAM, PHOTNU, FLAM, FNU, OBMAG, VEGAMAG])


# --------------- #
# Flux conversion #
# --------------- #

def spectral_density_vega(wav, vegaflux):
    """Flux equivalencies between PHOTLAM and VEGAMAG.

    Parameters
    ----------
    wav : `~astropy.units.quantity.Quantity`
        Quantity associated with values being converted
        (e.g., wavelength or frequency).

    vegaflux : `~astropy.units.quantity.Quantity`
        Flux of Vega at ``wav``.

    Returns
    -------
    eqv : list
        List of equivalencies.

    """
    vega_photlam = vegaflux.to(
        PHOTLAM, equivalencies=u.spectral_density(wav)).value

    def converter(x):
        return x / vega_photlam

    def iconverter(x):
        return x * vega_photlam

    return [(PHOTLAM, VEGAMAG.physical_unit, converter, iconverter)]


def spectral_density_count(wav, area):
    """Flux equivalencies between PHOTLAM and count/OBMAG.

    Parameters
    ----------
    wav : `~astropy.units.quantity.Quantity`
        Quantity associated with values being converted
        (e.g., wavelength or frequency).

    area : `~astropy.units.quantity.Quantity`
        Telescope collecting area.

    Returns
    -------
    eqv : list
        List of equivalencies.

    """
    from synphot.binning import calculate_bin_widths, calculate_bin_edges

    wav = wav.to(u.AA, equivalencies=u.spectral())
    area = area.to(AREA)
    bin_widths = calculate_bin_widths(calculate_bin_edges(wav))
    factor = bin_widths.value * area.value

    def converter_count(x):
        return x * factor

    def iconverter_count(x):
        return x / factor

    return [(PHOTLAM, u.count, converter_count, iconverter_count),
            (PHOTLAM, OBMAG.physical_unit, converter_count, iconverter_count)]


def convert_flux(wavelengths, fluxes, out_flux_unit, **kwargs):
    """Perform conversion for :ref:`supported flux units <synphot-flux-units>`.

    Parameters
    ----------
    wavelengths : array-like or `~astropy.units.quantity.Quantity`
        Wavelength values. If not a Quantity, assumed to be in
        Angstrom.

    fluxes : array-like or `~astropy.units.quantity.Quantity`
        Flux values. If not a Quantity, assumed to be in PHOTLAM.

    out_flux_unit : str or `~astropy.units.Unit`
        Output flux unit.

    area : float or `~astropy.units.quantity.Quantity`
        Area that fluxes cover. If not a Quantity, assumed to be in
        :math:`cm^{2}`. This value *must* be provided for conversions involving
        OBMAG and count, otherwise it is not needed.

    vegaspec : `~synphot.spectrum.SourceSpectrum`
        Vega spectrum that *must* be provided for conversions involving
        VEGAMAG, otherwise it is not needed. For instance, it can be
        obtained from :func:`~synphot.spectrum.SourceSpectrum.from_vega`.

    Returns
    -------
    out_flux : `~astropy.units.quantity.Quantity`
        Converted flux values.

    Raises
    ------
    astropy.units.UnitsError
        Conversion failed.

    synphot.exceptions.SynphotError
        Area or Vega spectrum is not given when needed.

    """
    if not isinstance(fluxes, u.Quantity):
        fluxes = fluxes * PHOTLAM

    out_flux_unit = validate_unit(out_flux_unit)
    out_flux_unit_name = out_flux_unit.to_string()
    in_flux_unit_name = fluxes.unit.to_string()

    # No conversion necessary
    if in_flux_unit_name == out_flux_unit_name:
        return fluxes

    in_flux_type = fluxes.unit.physical_type
    out_flux_type = out_flux_unit.physical_type

    # Wavelengths must Quantity
    if not isinstance(wavelengths, u.Quantity):
        wavelengths = wavelengths * u.AA

    eqv = u.spectral_density(wavelengths)

    # Use built-in astropy equivalencies
    try:
        out_flux = fluxes.to(out_flux_unit, eqv)

    # Use PHOTLAM as in-between unit
    except u.UnitConversionError:
        # Convert input unit to PHOTLAM
        if fluxes.unit == PHOTLAM:
            flux_photlam = fluxes
        elif in_flux_type != 'unknown':
            flux_photlam = fluxes.to(PHOTLAM, eqv)
        else:
            flux_photlam = _convert_flux(
                wavelengths, fluxes, PHOTLAM, **kwargs)

        # Convert PHOTLAM to output unit
        if out_flux_unit == PHOTLAM:
            out_flux = flux_photlam
        elif out_flux_type != 'unknown':
            out_flux = flux_photlam.to(out_flux_unit, eqv)
        else:
            out_flux = _convert_flux(
                wavelengths, flux_photlam, out_flux_unit, **kwargs)

    return out_flux


def _convert_flux(wavelengths, fluxes, out_flux_unit, area=None,
                  vegaspec=None):
    """Flux conversion for PHOTLAM <-> X."""
    flux_unit_names = (fluxes.unit.to_string(), out_flux_unit.to_string())

    if PHOTLAM.to_string() not in flux_unit_names:
        raise exceptions.SynphotError(
            'PHOTLAM must be one of the conversion units but get '
            '{0}.'.format(flux_unit_names))

    # VEGAMAG
    if VEGAMAG.to_string() in flux_unit_names:
        from synphot.spectrum import SourceSpectrum

        if not isinstance(vegaspec, SourceSpectrum):
            raise exceptions.SynphotError('Vega spectrum is missing.')

        flux_vega = vegaspec(wavelengths)

        out_flux = fluxes.to(
            out_flux_unit,
            equivalencies=spectral_density_vega(wavelengths, flux_vega))

    # OBMAG or count
    elif (u.count in (fluxes.unit, out_flux_unit) or
          OBMAG.to_string() in flux_unit_names):
        if area is None:
            raise exceptions.SynphotError(
                'Area is compulsory for conversion involving count or OBMAG.')
        elif not isinstance(area, u.Quantity):
            area = area * AREA

        out_flux = fluxes.to(
            out_flux_unit,
            equivalencies=spectral_density_count(wavelengths, area))

    else:
        raise u.UnitsError('{0} and {1} are not convertible'.format(
            fluxes.unit, out_flux_unit))

    return out_flux


# ----------------- #
# Utility functions #
# ----------------- #

def validate_unit(input_unit):
    """Validate unit.

    To be compatible with existing SYNPHOT data files:

        * 'angstroms' and 'inversemicrons' are accepted although
          unrecognized by astropy units
        * 'transmission', 'extinction', and 'emissivity' are
          converted to astropy dimensionless unit

    Parameters
    ----------
    input_unit : str or `~astropy.units.Unit`
        Unit to validate.

    Returns
    -------
    output_unit : `~astropy.units.Unit`
        Validated unit.

    Raises
    ------
    synphot.exceptions.SynphotError
        Invalid unit.

    """
    if isinstance(input_unit, str):
        input_unit_lowcase = input_unit.lower()

        # Backward-compatibility
        if input_unit_lowcase == 'angstroms':
            output_unit = u.AA
        elif input_unit_lowcase == 'inversemicrons':
            output_unit = u.micron ** -1
        elif input_unit_lowcase in ('transmission', 'extinction',
                                    'emissivity'):
            output_unit = THROUGHPUT
        elif input_unit_lowcase == 'jy':
            output_unit = u.Jy

        # Work around mag unit limitations
        elif input_unit_lowcase in ('stmag', 'mag(st)'):
            output_unit = u.STmag
        elif input_unit_lowcase in ('abmag', 'mag(ab)'):
            output_unit = u.ABmag
        elif input_unit_lowcase in ('obmag', 'mag(ob)'):
            output_unit = OBMAG
        elif input_unit_lowcase in ('vegamag', 'mag(vega)'):
            output_unit = VEGAMAG

        else:
            try:  # astropy.units is case-sensitive
                output_unit = u.Unit(input_unit)
            except ValueError:  # synphot is case-insensitive
                output_unit = u.Unit(input_unit_lowcase)

    elif isinstance(input_unit, (u.UnitBase, u.LogUnit)):
        output_unit = input_unit

    else:
        raise exceptions.SynphotError(
            '{0} must be a recognized string or '
            'astropy.units.Unit'.format(input_unit))

    return output_unit


def validate_wave_unit(wave_unit):
    """Like :func:`validate_unit` but specific to wavelength."""
    output_unit = validate_unit(wave_unit)
    unit_type = output_unit.physical_type

    if unit_type not in ('length', 'wavenumber', 'frequency'):
        raise exceptions.SynphotError(
            'wavelength physical type is not length, wave number, or '
            'frequency: {0}'.format(unit_type))

    return output_unit


def validate_quantity(input_value, output_unit, equivalencies=[]):
    """Validate quantity (value and unit).

    .. note::

        For flux conversion, use :func:`convert_flux` instead.

    Parameters
    ----------
    input_value : number, array-like, or `~astropy.units.quantity.Quantity`
        Quantity to validate. If not a Quantity, assumed to be
        already in output unit.

    output_unit : str or `~astropy.units.Unit`
        Output quantity unit.

    equivalencies : list of equivalence pairs, optional
        See `astropy.units`.

    Returns
    -------
    output_value : `~astropy.units.quantity.Quantity`
        Validated quantity in given unit.

    """
    output_unit = validate_unit(output_unit)

    if isinstance(input_value, u.Quantity):
        output_value = input_value.to(output_unit, equivalencies=equivalencies)
    else:
        output_value = input_value * output_unit

    return output_value



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test models.py module.

.. note::

    ``GaussianAbsorption1D`` and ``RedshiftScaleFactor`` are already being
    tested within existing Astropy PRs.

    ``get_waveset()`` is tested in test_spectrum.py.

"""

# STDLIB
import os
import warnings

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.modeling.models import Const1D
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning

# LOCAL
from synphot import specio, units
from synphot.models import (BlackBody1D, ConstFlux1D, Empirical1D,
                            PowerLawFlux1D, get_metadata)


def setup_module(module):
    import astropy.constants as const
    from astropy.constants import si, astropyconst13

    const.sigma_sb = si.sigma_sb = astropyconst13.sigma_sb
    const.h = si.h = astropyconst13.h
    const.k_B = si.k_B = astropyconst13.k_B


def teardown_module(module):
    import astropy.constants as const
    from astropy.constants import si, astropyconst20

    const.sigma_sb = si.sigma_sb = astropyconst20.sigma_sb
    const.h = si.h = astropyconst20.h
    const.k_B = si.k_B = astropyconst20.k_B


class TestBlackBody1D:
    """Test BlackBody1D model."""
    def setup_class(self):
        self.m1 = BlackBody1D(temperature=5500)

    def test_lambda_max(self):
        np.testing.assert_allclose(self.m1.lambda_max, 5268.67, rtol=1e-5)

    def test_sampleset(self):
        f1 = self.m1(self.m1.sampleset())
        assert f1[0] == 0
        assert f1[-1] < self.m1(self.m1.lambda_max) * 0.05

    def test_eval(self):
        np.testing.assert_allclose(
            self.m1(np.arange(3000, 3100, 10)),
            [1.20906423e+17, 1.22815123e+17, 1.24735543e+17, 1.26667499e+17,
             1.28610806e+17, 1.30565276e+17, 1.32530722e+17, 1.34506953e+17,
             1.36493780e+17, 1.38491010e+17])

    def test_multi_n_models(self):
        m2 = BlackBody1D(temperature=[100, 10000], n_models=2)
        np.testing.assert_allclose(
            m2.lambda_max, [2.8977685e5, 2897.7685], rtol=1e-5)
        np.testing.assert_allclose(
            m2(20000), [2.14331496e-14, 3.55819086e+17])


class TestConstFlux1D:
    """Test ConstFlux1D model."""
    def setup_class(self):
        self.w = np.arange(1, 21000, 5000)

    @pytest.mark.parametrize(
        'amplitude',
        [1, 1 * units.PHOTNU, 1 * units.FLAM, 1 * units.FNU,
         1 * u.Jy, 1 * u.mJy])
    def test_linear(self, amplitude):
        if isinstance(amplitude, u.Quantity):
            ans = amplitude.value
            flux_unit = amplitude.unit
        else:
            ans = amplitude
            flux_unit = units.PHOTLAM

        m = ConstFlux1D(amplitude=amplitude)
        f = units.convert_flux(self.w, m(self.w) * units.PHOTLAM, flux_unit)

        assert m._flux_unit == flux_unit
        np.testing.assert_allclose(f.value, ans)

    @pytest.mark.parametrize(
        ('in_unit', 'out_unit', 'val'),
        [(u.STmag, units.FLAM, 3.63e-9),
         (u.ABmag, units.FNU, 3.63e-20)])
    def test_mag(self, in_unit, out_unit, val):
        m = ConstFlux1D(amplitude=0 * in_unit)
        f = units.convert_flux(self.w, m(self.w) * units.PHOTLAM, out_unit)
        np.testing.assert_allclose(f.value, val, rtol=2.5e-4)

    def test_multi_n_models(self):
        m = ConstFlux1D(amplitude=[1, 2], n_models=2)
        np.testing.assert_array_equal(m(1000), [1, 2])

    @pytest.mark.parametrize(
        'flux_unit', [u.count, units.OBMAG, units.VEGAMAG, u.AA])
    def test_invalid_units(self, flux_unit):
        with pytest.raises(NotImplementedError):
            ConstFlux1D(amplitude=1 * flux_unit)


class TestEmpirical1D:
    """Test Empirical1D model."""
    def setup_class(self):
        filename = get_pkg_data_filename(
            os.path.join('data', 'hst_acs_hrc_f555w_x_grw70d5824.fits'),
            package='synphot.tests')
        hdr, x, f = specio.read_spec(filename)
        y = units.convert_flux(x, f, units.PHOTLAM)
        self.flux_flam = f.value
        self.w = x.value
        self.m = Empirical1D(points=self.w, lookup_table=y.value)

    def test_sampleset(self):
        np.testing.assert_array_equal(self.m.sampleset(), self.w)

    def test_eval(self):
        # Sample at existing wavelength (no interpolation)
        x = self.w[5000]
        y = units.convert_flux(x, self.m(x), units.FLAM)
        np.testing.assert_allclose(y.value, self.flux_flam[5000])

        # Sampling with interpolation
        w = [4956.8, 4959.55, 4962.3]
        y = units.convert_flux(w, self.m(w), units.FLAM)
        np.testing.assert_allclose(
            y.value, [8.57166622e-15, 8.86174843e-15, 8.68707743e-15],
            rtol=1e-6)

        # Descending order
        w2 = w[::-1]
        f2 = self.m(w2)
        y = units.convert_flux(w2, f2, units.FLAM)
        np.testing.assert_allclose(
            y.value, [8.68707743e-15, 8.86174843e-15, 8.57166622e-15],
            rtol=1e-6)

        # New model with descending order
        m2 = Empirical1D(points=w2, lookup_table=f2)
        y = units.convert_flux(w, m2(w), units.FLAM)
        np.testing.assert_allclose(
            y.value, [8.57166622e-15, 8.86174843e-15, 8.68707743e-15],
            rtol=1e-6)

    @pytest.mark.parametrize(
        ('keep_neg', 'ans'),
        [(True, [-1.1, 0, 1.1]),
         (False, [0, 0, 1.1])])
    def test_neg_array(self, keep_neg, ans):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', message=r'.*contained negative flux or throughput.*',
                category=AstropyUserWarning)
            m2 = Empirical1D(points=[1, 2, 3], lookup_table=[-1.1, 0, 1.1],
                             keep_neg=keep_neg)
        np.testing.assert_array_equal(m2([1, 2, 3]), ans)
        if not keep_neg:
            assert 'NegativeFlux' in m2.meta['warnings']

    @pytest.mark.parametrize(
        ('keep_neg', 'ans'),
        [(True, -1),
         (False, 0)])
    def test_neg_scalar(self, keep_neg, ans):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', message=r'.*contained negative flux or throughput.*',
                category=AstropyUserWarning)
            m2 = Empirical1D(points=[1, 2, 3], lookup_table=[-1, 0, 1],
                             keep_neg=keep_neg)
        np.testing.assert_array_equal(m2(1), ans)
        if not keep_neg:
            assert 'NegativeFlux' in m2.meta['warnings']

    @pytest.mark.parametrize(
        ('tab', 'ans'),
        [([0, 1, 0], True),
         ([0, 1, 1], False)])
    def test_taper(self, tab, ans):
        m2 = Empirical1D(points=[1, 2, 3], lookup_table=tab)
        assert m2.is_tapered() is ans

    def test_extrap(self):
        """Test extrapolation of constant at both ends, as done in
        ASTROLIB PYSYNPHOT.
        """
        m2 = Empirical1D(
            points=[1000, 2000, 3000, 4000],
            lookup_table=[0.01, 5.0, 10.6, 1.5])
        assert m2(900) == 0.01
        assert m2(10000) == 1.5
        np.testing.assert_allclose(
            m2([900, 1000, 1500, 10000]), [0.01, 0.01, 2.505, 1.5])


class TestPowerLawFlux1D:
    """Test PowerLawFlux1D model."""
    def setup_class(self):
        self.w = np.arange(3000, 3100, 10)
        self.m = PowerLawFlux1D(amplitude=1, x_0=6000, alpha=4)

    def test_eval(self):
        assert self.m._flux_unit == units.PHOTLAM
        np.testing.assert_allclose(
            self.m(self.w),
            [16, 15.78843266, 15.58035072, 15.37568551, 15.17436992,
             14.97633838, 14.78152682, 14.5898726, 14.40131453, 14.21579277],
            rtol=1e-6)

    def test_normalization(self):
        assert self.m(self.m.x_0) == 1

    def test_multi_n_models(self):
        m2 = PowerLawFlux1D(
            amplitude=[1, 1] * units.FLAM, x_0=[0.3, 0.305] * u.micron,
            alpha=[4, 1], n_models=2)
        y = units.convert_flux(
            self.w, m2(self.w, model_set_axis=False) * units.PHOTLAM,
            units.FLAM)
        ans = [[1, 0.98677704, 0.97377192, 0.96098034, 0.94839812,
                0.93602115, 0.92384543, 0.91186704, 0.90008216, 0.88848705],
               [1.01666667, 1.01328904, 1.00993377, 1.00660066, 1.00328947,
                1, 0.99673203, 0.99348534, 0.99025974, 0.98705502]]
        np.testing.assert_allclose(y.value, ans, rtol=1e-6)

    @pytest.mark.parametrize('flux_unit', [u.STmag, u.ABmag])
    def test_valid_mag(self, flux_unit):
        m = PowerLawFlux1D(amplitude=1 * flux_unit, x_0=6000, alpha=4)
        assert m.amplitude == 1
        assert m._flux_unit == flux_unit

    @pytest.mark.parametrize(
        'flux_unit', [u.count, units.OBMAG, units.VEGAMAG, u.AA])
    def test_invalid_units(self, flux_unit):
        with pytest.raises(NotImplementedError):
            PowerLawFlux1D(amplitude=1 * flux_unit, x_0=5000, alpha=4)


def test_get_metadata():
    m1 = Const1D()
    m1.meta['description'] = 'a constant'
    m1.meta['foo'] = 42.0

    m2 = Const1D()
    m2.meta['description'] = 'another constant'

    m3 = Const1D()
    m3.meta[42] = 'answer'

    m = (m1 + m2) * m3
    meta = get_metadata(m)
    keys = list(meta.keys())
    assert len(keys) == 3
    assert meta['description'] == 'another constant'
    assert meta['foo'] == 42
    assert meta[42] == 'answer'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Synphot configurable items.

The default configuration heavily depends on STScI TRDS structure
but it can be easily re-configured as the user wishes via
`astropy.config`.

"""

# ASTROPY
from astropy.config import ConfigNamespace, ConfigItem

__all__ = ['conf']


class Conf(ConfigNamespace):
    """Configuration parameters."""

    # First option in list is the default.
    default_integrator = ConfigItem(
        ['trapezoid', 'analytical'], 'Default integrator to use')

    # STANDARD STARS
    vega_file = ConfigItem(
        'https://ssb.stsci.edu/trds/calspec/alpha_lyr_stis_010.fits', 'Vega')

    # REDDENING/EXTINCTION LAWS
    lmc30dor_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/lmc_30dorshell_001.fits',
        'Gordon et al. 2003, ApJ, 594, 279; R_V = 2.76')
    lmcavg_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/lmc_diffuse_001.fits',
        'Gordon et al. 2003, ApJ, 594, 279; R_V = 3.41')
    mwavg_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/milkyway_diffuse_001.fits',
        'Cardelli, Clayton, & Mathis 1989, ApJ, 345, 245; R_V = 3.10')
    mwdense_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/milkyway_dense_001.fits',
        'Cardelli, Clayton, & Mathis 1989, ApJ, 345, 245; R_V = 5.00')
    mwrv21_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/milkyway_rv21_001.fits',
        'Cardelli, Clayton, & Mathis 1989, ApJ, 345, 245; R_V = 2.1')
    mwrv40_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/milkyway_rv4_001.fits',
        'Cardelli, Clayton, & Mathis 1989, ApJ, 345, 245; R_V = 4.0')
    smcbar_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/smc_bar_001.fits',
        'Gordon et al. 2003, ApJ, 594, 279; R_V=2.74')
    xgal_file = ConfigItem(
        'https://ssb.stsci.edu/trds/extinction/xgal_starburst_001.fits',
        'Calzetti et al. 2000, ApJ, 533, 682')

    # COMMON FILTER BANDPASS
    bessel_h_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/bessell_h_004_syn.fits',
        'Bessel H')
    bessel_j_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/bessell_j_003_syn.fits',
        'Bessel J')
    bessel_k_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/bessell_k_003_syn.fits',
        'Bessel K')
    cousins_i_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/cousins_i_004_syn.fits',
        'Cousins I')
    cousins_r_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/cousins_r_004_syn.fits',
        'Cousins R')
    johnson_b_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/johnson_b_004_syn.fits',
        'Johnson B')
    johnson_i_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/johnson_i_003_syn.fits',
        'Johnson I')
    johnson_j_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/johnson_j_003_syn.fits',
        'Johnson J')
    johnson_k_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/johnson_k_003_syn.fits',
        'Johnson K')
    johnson_r_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/johnson_r_003_syn.fits',
        'Johnson R')
    johnson_u_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/johnson_u_004_syn.fits',
        'Johnson U')
    johnson_v_file = ConfigItem(
        'https://ssb.stsci.edu/trds/comp/nonhst/johnson_v_004_syn.fits',
        'Johnson V')


conf = Conf()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test units.py module.

.. note:: VEGAMAG conversion is tested in test_spectrum_source.py.

.. note:: spectral_density_integrated is tested in astropy>=4.1.

"""

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.tests.helper import assert_quantity_allclose

# LOCAL
from synphot import exceptions, units

# Wavelength conversions
_wave_angstrom = [0.1, 5000.0, 10000.0] * u.AA
_wavenum_micron = [1e+5, 2.0, 1.0] * (u.micron ** -1)
_freq = [2.99792458e+19, 5.99584916e+14, 2.99792458e+14] * u.Hz

# Flux conversions
# http://ssb.stsci.edu/cdbs/supplemental_calspec/grw_70d5824_stisnic_002.ascii
_area = 45238.93416 * units.AREA  # HST
_wave = [4956.8, 4959.55, 4962.3] * u.AA
_flux_photlam = [9.7654e-3, 1.003896e-2, 9.78473e-3] * units.PHOTLAM
_flux_photnu = [8.00335589e-14, 8.23668949e-14, 8.03700310e-14] * units.PHOTNU
_flux_flam = [3.9135e-14, 4.0209e-14, 3.9169e-14] * units.FLAM
_flux_fnu = [3.20735792e-25, 3.29903646e-25, 3.21727226e-25] * units.FNU
_flux_jy = [3.20735792e-2, 3.29903646e-2, 3.21727226e-2] * u.Jy
_flux_count = [1214.88479883, 1248.91795446, 1217.28946691] * u.count
_flux_stmag = [12.41858665, 12.38919182, 12.41764379] * u.STmag
_flux_abmag = [12.63463143, 12.60403221, 12.63128047] * u.ABmag
_flux_obmag = [-7.71133775, -7.74133477, -7.71348466] * units.OBMAG
_flux_vegamag = [12.72810665, 12.69861694, 12.72605148] * units.VEGAMAG


def test_implicit_assumptions():
    """These assumptions must be valid for proper conversions."""
    assert units.HC.unit == u.AA * u.erg
    assert units.AREA.physical_type == 'area'
    assert units.THROUGHPUT.physical_type == 'dimensionless'
    np.testing.assert_allclose(units.SR_PER_ARCSEC2, 2.3504430539097885e-11)


@pytest.mark.parametrize(
    ('in_u', 'out_u'),
    [('angstroms', u.AA),
     ('inversemicrons', u.micron ** -1),
     ('transmission', units.THROUGHPUT),
     ('TRANSMISSION', units.THROUGHPUT),
     ('extinction', units.THROUGHPUT),
     ('emissivity', units.THROUGHPUT),
     ('photlam', units.PHOTLAM),
     ('photnu', units.PHOTNU),
     ('flam', units.FLAM),
     ('fnu', units.FNU),
     ('stmag', u.STmag),
     ('abmag', u.ABmag),
     ('obmag', units.OBMAG),
     ('vegamag', units.VEGAMAG),
     ('Kelvin', u.K),
     (u.m, u.m)])
def test_validate_unit(in_u, out_u):
    """Test unit validation."""
    assert units.validate_unit(in_u) == out_u


@pytest.mark.parametrize(
    ('in_u', 'out_u'),
    [('angstroms', u.AA),
     ('inversemicrons', u.micron ** -1),
     ('Hz', u.Hz)])
def test_validate_wave_unit(in_u, out_u):
    """Test wavelength unit validation."""
    assert units.validate_wave_unit(in_u) == out_u


def test_validate_unit_exceptions():
    """Test that unit validation raises appropriate exceptions."""
    with pytest.raises(exceptions.SynphotError):
        units.validate_unit(10)
    with pytest.raises(ValueError):
        units.validate_unit('foo')
    with pytest.raises(exceptions.SynphotError):
        units.validate_wave_unit('Kelvin')


@pytest.mark.parametrize(
    ('in_val', 'out_u', 'eqv', 'ans'),
    [(100.0, units.AREA, [], 100.0),
     (100.0 * units.AREA, u.m * u.m, [], 0.01),
     (_wave_angstrom, u.micron ** -1, u.spectral(), _wavenum_micron.value)])
def test_validate_quantity(in_val, out_u, eqv, ans):
    """Test quantity validation."""
    result = units.validate_quantity(in_val, out_u, equivalencies=eqv)
    np.testing.assert_allclose(result.value, ans)
    assert result.unit == out_u


@pytest.mark.parametrize(
    ('in_q', 'out_u', 'ans'),
    [(_wave_angstrom, u.Hz, _freq),
     (_freq, u.AA, _wave_angstrom),
     (_wave_angstrom, u.micron ** -1, _wavenum_micron),
     (_wavenum_micron, u.AA, _wave_angstrom),
     (_freq, u.micron ** -1, _wavenum_micron),
     (_wavenum_micron, u.Hz, _freq)])
def test_wave_conversion(in_q, out_u, ans):
    """Full equivalencies test with direct conversion."""
    result = in_q.to(out_u, equivalencies=u.spectral())
    np.testing.assert_allclose(result.value, ans.value)
    assert result.unit == ans.unit


@pytest.mark.parametrize(
    ('in_q', 'out_u', 'ans', 'support_scalar'),
    [(_flux_photlam.value, units.PHOTLAM, _flux_photlam, True),
     (_flux_photlam, u.count, _flux_count, False),
     (_flux_count, units.PHOTLAM, _flux_photlam, False),
     (_flux_photlam, units.OBMAG, _flux_obmag, False),
     (_flux_obmag, units.PHOTLAM, _flux_photlam, False),
     (_flux_count, units.OBMAG, _flux_obmag, False),
     (_flux_obmag, u.count, _flux_count, False),
     (_flux_photlam, units.FLAM, _flux_flam, True),
     (_flux_flam, units.PHOTLAM, _flux_photlam, True),
     (_flux_photlam, u.STmag, _flux_stmag, True),
     (_flux_stmag, units.PHOTLAM, _flux_photlam, True),
     (_flux_flam, u.STmag, _flux_stmag, True),
     (_flux_stmag, units.FLAM, _flux_flam, True),
     (_flux_photlam, units.PHOTNU, _flux_photnu, True),
     (_flux_photnu, units.PHOTLAM, _flux_photlam, True),
     (_flux_photlam, units.FNU, _flux_fnu, True),
     (_flux_fnu, units.PHOTLAM, _flux_photlam, True),
     (_flux_photlam, u.ABmag, _flux_abmag, True),
     (_flux_abmag, units.PHOTLAM, _flux_photlam, True),
     (_flux_fnu, u.ABmag, _flux_abmag, True),
     (_flux_abmag, units.FNU, _flux_fnu, True),
     (_flux_fnu, u.STmag, _flux_stmag, True),
     (_flux_fnu, u.mJy, _flux_jy.to(u.mJy), True),
     (_flux_photlam, u.Jy, _flux_jy, True),
     (_flux_jy, units.PHOTLAM, _flux_photlam, True),
     (_flux_flam, u.Jy, _flux_jy, True),
     (np.zeros(3) * units.FNU, units.FLAM, np.zeros(3) * units.FLAM, True)])
def test_flux_conversion(in_q, out_u, ans, support_scalar):
    """Test flux conversion, except VEGAMAG."""
    result = units.convert_flux(_wave, in_q, out_u, area=_area)
    np.testing.assert_allclose(result.value, ans.value, rtol=1e-6)
    assert result.unit == ans.unit

    # Scalar should work, except for count and OBMAG, which need bin centers.
    if support_scalar:
        i = 0
        result = units.convert_flux(_wave[i], in_q[i], out_u, area=_area)
        np.testing.assert_allclose(result.value, ans[i].value, rtol=1e-6)
        assert result.unit == ans[i].unit


def test_flux_conversion_exceptions():
    """Test for appropriate exceptions."""
    # Invalid flux unit
    with pytest.raises(u.UnitsError):
        units.convert_flux(_wave, _wave, units.PHOTLAM)
    with pytest.raises(u.UnitsError):
        units.convert_flux(_wave, _flux_photlam, u.AA)

    # Missing Vega spectrum
    with pytest.raises(exceptions.SynphotError):
        units.convert_flux(_wave, _flux_fnu, units.VEGAMAG, vegaspec=None)

    # Missing area
    with pytest.raises(exceptions.SynphotError):
        units.convert_flux(_wave, _flux_photlam, u.count, area=None)
    with pytest.raises(exceptions.SynphotError):
        units.convert_flux(_wave, _flux_obmag, units.PHOTLAM, area=None)


def test_vegamag_obmag_calculations():
    assert_quantity_allclose(
        5 * units.VEGAMAG - 2.5 * units.VEGAMAG, u.Magnitude(2.5))
    assert_quantity_allclose(
        (5 * units.VEGAMAG - 2.5 * units.VEGAMAG).to(u.one), 0.1)

    # Should not be interchangeable with astropy mag unit or with another
    # custom mag unit, but error is only raised if .to(u.one) is called.
    msg = 'subtract magnitudes so the unit got lost'
    with pytest.raises(u.UnitConversionError, match=msg):
        (5 * units.VEGAMAG - 2.5 * u.STmag).to(u.one)
    with pytest.raises(u.UnitConversionError, match=msg):
        5 * units.VEGAMAG - 2.5 * units.OBMAG.to(u.one)



import pytest
from astropy import units as u
from astropy.table import Table
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename

from synphot.filter_parameterization.filter_fft import (
    filter_from_fft, analytical_model_from_fft)


class TestSVOFilters:
    """Unit test adapted from tynt package."""
    def setup_class(self):
        self.fft_table = Table.read(
            get_pkg_data_filename('data/fft_test_data.fits'))
        self.fft_table.add_index('filter')
        self.fft_cols = self.fft_table.colnames[5:]
        self.lambda_unit = self.fft_table['lambda_0'].unit
        self.dlamb_unit = self.fft_table['delta_lambda'].unit

    def inputs_from_table(self, filtername):
        row = self.fft_table.loc[filtername]
        n_lambda = row['n_lambda']
        lambda_0 = row['lambda_0'] * self.lambda_unit
        delta_lambda = row['delta_lambda'] * self.dlamb_unit
        tr_max = row['tr_max']
        fft_pars = list(row[self.fft_cols])
        return n_lambda, lambda_0, delta_lambda, tr_max, fft_pars

    # Answers from:
    # http://svo2.cab.inta-csic.es/theory/fps/index.php?mode=browse&gname=SLOAN
    # http://svo2.cab.inta-csic.es/theory/fps/index.php?mode=browse&gname=2MASS
    # http://svo2.cab.inta-csic.es/theory/fps/index.php?mode=browse&gname=Generic&gname2=Johnson
    @pytest.mark.parametrize(
        ('filtername', 'lambda_mean_true', 'w_eff_true'),
        [('SLOAN/SDSS.u', 3561.8 * u.AA, 558.4 * u.AA),
         ('SLOAN/SDSS.g', 4718.9 * u.AA, 1158.4 * u.AA),
         ('SLOAN/SDSS.r', 6185.2 * u.AA, 1111.2 * u.AA),
         ('SLOAN/SDSS.i', 7499.7 * u.AA, 1044.6 * u.AA),
         ('SLOAN/SDSS.z', 8961.5 * u.AA, 1124.6 * u.AA),
         ('2MASS/2MASS.J', 12350 * u.AA, 1624.1 * u.AA),
         ('2MASS/2MASS.H', 16620 * u.AA, 2509.4 * u.AA),
         ('2MASS/2MASS.Ks', 21590 * u.AA, 2618.9 * u.AA),
         ('Generic/Johnson.U', 3531.1 * u.AA, 657 * u.AA),
         ('Generic/Johnson.B', 4430.4 * u.AA, 972.7 * u.AA),
         ('Generic/Johnson.V', 5537.2 * u.AA, 889.7 * u.AA),
         ('Generic/Johnson.R', 6939.6 * u.AA, 2070 * u.AA),
         ('Generic/Johnson.I', 8780.7 * u.AA, 2316 * u.AA)])
    def test_lambda_eff_w_eff(self, filtername, lambda_mean_true, w_eff_true):
        inputs = self.inputs_from_table(filtername)
        bp = filter_from_fft(*inputs)
        assert_quantity_allclose(bp.avgwave(), lambda_mean_true, rtol=0.03)
        assert_quantity_allclose(bp.rectwidth(), w_eff_true, rtol=0.1)

    def test_astropy_model(self):
        inputs = self.inputs_from_table('SLOAN/SDSS.r')
        bp = filter_from_fft(*inputs)
        m = analytical_model_from_fft(*inputs)
        wave = bp.waveset
        assert_quantity_allclose(bp(wave), m(wave))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Synthetic photometry utility functions."""

# STDLIB
import os
from shutil import copyfile

# THIRD-PARTY
import numpy as np

# ASTROPY
from astropy import units as u
from astropy.config import ConfigItem
from astropy.utils.data import download_file

# LOCAL
from synphot import exceptions, units

__all__ = ['overlap_status', 'validate_totalflux', 'validate_wavelengths',
           'generate_wavelengths', 'merge_wavelengths', 'download_data']


def overlap_status(a, b):
    """Check overlap between two arrays.

    Parameters
    ----------
    a, b : array-like
        Arrays to check. Assumed to be in the same unit.

    Returns
    -------
    result : {'full', 'partial', 'none'}
        * 'full' - ``a`` is within or same as ``b``
        * 'partial' - ``a`` partially overlaps with ``b``
        * 'none' - ``a`` does not overlap ``b``

    """
    # Get the endpoints
    a1, a2 = a.min(), a.max()
    b1, b2 = b.min(), b.max()

    # Do the comparison
    if a1 >= b1 and a2 <= b2:
        result = 'full'
    elif a2 < b1 or b2 < a1:
        result = 'none'
    else:
        result = 'partial'

    return result


def validate_totalflux(totalflux):
    """Check integrated flux for invalid values.

    Parameters
    ----------
    totalflux : float
        Integrated flux.

    Raises
    ------
    synphot.exceptions.SynphotError
        Input is zero, negative, or not a number.

    """
    if totalflux <= 0.0:
        raise exceptions.SynphotError('Integrated flux is <= 0')
    elif np.isnan(totalflux):
        raise exceptions.SynphotError('Integrated flux is NaN')
    elif np.isinf(totalflux):
        raise exceptions.SynphotError('Integrated flux is infinite')


def validate_wavelengths(wavelengths):
    """Check wavelengths for ``synphot`` compatibility.

    Wavelengths must satisfy these conditions:

        * valid unit type, if given
        * no zeroes
        * monotonic ascending or descending
        * no duplicate values

    Parameters
    ----------
    wavelengths : array-like or `~astropy.units.quantity.Quantity`
        Wavelength values.

    Raises
    ------
    synphot.exceptions.SynphotError
        Wavelengths unit type is invalid.

    synphot.exceptions.DuplicateWavelength
        Wavelength array contains duplicate entries.

    synphot.exceptions.UnsortedWavelength
        Wavelength array is not monotonic.

    synphot.exceptions.ZeroWavelength
        Negative or zero wavelength occurs in wavelength array.

    """
    if isinstance(wavelengths, u.Quantity):
        units.validate_wave_unit(wavelengths.unit)
        wave = wavelengths.value
    else:
        wave = wavelengths

    if np.isscalar(wave):
        wave = [wave]

    wave = np.asarray(wave)

    # Check for zeroes
    if np.any(wave <= 0):
        raise exceptions.ZeroWavelength(
            'Negative or zero wavelength occurs in wavelength array',
            rows=np.where(wave <= 0)[0])

    # Check for monotonicity
    sorted_wave = np.sort(wave)
    if not np.alltrue(sorted_wave == wave):
        if np.alltrue(sorted_wave[::-1] == wave):
            pass  # Monotonic descending is allowed
        else:
            raise exceptions.UnsortedWavelength(
                'Wavelength array is not monotonic',
                rows=np.where(sorted_wave != wave)[0])

    # Check for duplicate values
    if wave.size > 1:
        dw = sorted_wave[1:] - sorted_wave[:-1]
        if np.any(dw == 0):
            raise exceptions.DuplicateWavelength(
                'Wavelength array contains duplicate entries',
                rows=np.where(dw == 0)[0])


def generate_wavelengths(minwave=500, maxwave=26000, num=10000, delta=None,
                         log=True, wave_unit=u.AA):
    """Generate wavelength array to be used for spectrum sampling.

    .. math::

        minwave \\le \\lambda < maxwave

    Parameters
    ----------
    minwave, maxwave : float
        Lower and upper limits of the wavelengths.
        These must be values in linear space regardless of ``log``.

    num : int
        The number of wavelength values.
        This is only used when ``delta=None``.

    delta : float or `None`
        Delta between wavelength values.
        When ``log=True``, this is the spacing in log space.

    log : bool
        If `True`, the wavelength values are evenly spaced in log scale.
        Otherwise, spacing is linear.

    wave_unit : str or `~astropy.units.Unit`
        Wavelength unit. Default is Angstrom.

    Returns
    -------
    waveset : `~astropy.units.quantity.Quantity`
        Generated wavelength set.

    waveset_str : str
        Info string associated with the result.

    """
    wave_unit = units.validate_unit(wave_unit)

    if delta is not None:
        num = None

    waveset_str = 'Min: {0}, Max: {1}, Num: {2}, Delta: {3}, Log: {4}'.format(
        minwave, maxwave, num, delta, log)

    # Log space
    if log:
        logmin = np.log10(minwave)
        logmax = np.log10(maxwave)

        if delta is None:
            waveset = np.logspace(logmin, logmax, num, endpoint=False)
        else:
            waveset = 10 ** np.arange(logmin, logmax, delta)

    # Linear space
    else:
        if delta is None:
            waveset = np.linspace(minwave, maxwave, num, endpoint=False)
        else:
            waveset = np.arange(minwave, maxwave, delta)

    return waveset.astype(np.float64) * wave_unit, waveset_str


def merge_wavelengths(waveset1, waveset2, threshold=1e-12):
    """Return the union of the two sets of wavelengths using
    :func:`numpy.union1d`.

    The merged wavelengths may sometimes contain numbers which are nearly
    equal but differ at levels as small as 1e-14. Having values this
    close together can cause problems down the line. So, here we test
    whether any such small differences are present, with a small
    difference defined as less than ``threshold``. If a small
    difference is present, the lower of the too-close pair is removed.

    Parameters
    ----------
    waveset1, waveset2 : array-like or `None`
        Wavelength values, assumed to be in the same unit already.
        Also see :func:`~synphot.models.get_waveset`.

    threshold : float, optional
        Merged wavelength values are considered "too close together"
        when the difference is smaller than this number.
        The default is 1e-12.

    Returns
    -------
    out_wavelengths : array-like or `None`
        Merged wavelengths. `None` if undefined.

    """
    if waveset1 is None and waveset2 is None:
        out_wavelengths = None
    elif waveset1 is not None and waveset2 is None:
        out_wavelengths = waveset1
    elif waveset1 is None and waveset2 is not None:
        out_wavelengths = waveset2
    else:
        out_wavelengths = np.union1d(waveset1, waveset2)
        delta = out_wavelengths[1:] - out_wavelengths[:-1]
        i_good = np.where(delta > threshold)

        # Remove "too close together" duplicates
        if len(i_good[0]) < delta.size:
            out_wavelengths = np.append(
                out_wavelengths[i_good], out_wavelengths[-1])

    return out_wavelengths


def download_data(path_root, verbose=True, dry_run=False):
    """Download ``synphot`` data files to given root directory or
    the ``astropy`` cache.
    Download is skipped if a data file already exists.

    .. warning::

        Downloading data to ``astropy`` cache only is not recommended
        if you plan to provide a custom ``synphot.cfg``.

    Parameters
    ----------
    path_root : str or `None`
        Root directory for data files. If `None`, download to the ``astropy``
        cache location instead of a specific directory.

    verbose : bool
        Print extra information to screen.

    dry_run : bool
        Go through the logic but skip the actual download.
        This would return a list of files that *would have been*
        downloaded without network calls. The sub-directories
        would still be created regardless.
        Use this option for debugging or testing.

    Raises
    ------
    OSError
        Problem with directory.

    Returns
    -------
    file_list : list of str
        A list of downloaded files.

    """
    from synphot.config import conf  # Avoid potential circular import
    BASE_HOST = 'https://ssb.stsci.edu/trds/'

    if path_root is not None:
        if not os.path.exists(path_root):
            os.makedirs(path_root, exist_ok=True)
            if verbose:  # pragma: no cover
                print('Created {}'.format(path_root))
        elif not os.path.isdir(path_root):
            raise OSError('{} must be a directory'.format(path_root))

        if not path_root.endswith(os.sep):
            path_root += os.sep

    file_list = []

    # See https://github.com/astropy/astropy/issues/8524
    for cfgitem in conf.__class__.__dict__.values():
        if (not isinstance(cfgitem, ConfigItem) or
                not cfgitem.name.endswith('file')):
            continue

        url = cfgitem.defaultvalue

        if not url.startswith(BASE_HOST):
            if verbose:  # pragma: no cover
                print('{} is not from {}, skipping download'.format(
                    url, BASE_HOST))
            continue

        if path_root is not None:
            dst = url.replace(BASE_HOST, path_root).replace('/', os.sep)

            if os.path.exists(dst):
                if verbose:  # pragma: no cover
                    print('{} already exists, skipping download'.format(dst))
                continue

            # Create sub-directories, if needed.
            subdirs = os.path.dirname(dst)
            os.makedirs(subdirs, exist_ok=True)

        if not dry_run:  # pragma: no cover
            try:
                src = download_file(url, cache=True)
                if path_root is not None:
                    copyfile(src, dst)
            except Exception as exc:
                print('Download failed - {}'.format(str(exc)))
                continue

        if path_root is None:
            if dry_run:
                file_list.append(url)
            else:  # pragma: no cover
                file_list.append(src)
        else:
            file_list.append(dst)

        if verbose:  # pragma: no cover
            print('{} downloaded to {}'.format(url, file_list[-1]))

    return file_list



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module defines reddening laws and extinction curves."""

# STDLIB
import numbers

# THIRD-PARTY
import numpy as np
from astropy import units as u

# LOCAL
from synphot import exceptions, specio, units
from synphot.compat import HAS_DUST_EXTINCTION
from synphot.config import Conf
from synphot.models import Empirical1D
from synphot.spectrum import BaseUnitlessSpectrum

__all__ = ['ExtinctionModel1D', 'ReddeningLaw', 'ExtinctionCurve',
           'etau_madau']


class ExtinctionModel1D(Empirical1D):
    """Model to handle extinction curve.
    This is like :class:`~synphot.models.Empirical1D` except that
    its ``sampleset`` will not be propagated to composite spectrum.
    """
    def sampleset(self):
        """This simply returns `None`. Use ``numpy.squeeze(self.points)``
        instead for array (in Angstrom) that samples the model.
        """
        return None


class ReddeningLaw(BaseUnitlessSpectrum):
    """Class to handle reddening law.

    Parameters
    ----------
    modelclass, kwargs
        See `~synphot.spectrum.BaseSpectrum`.

    """
    def extinction_curve(self, ebv, wavelengths=None):
        """Generate extinction curve.

        .. math::

            A(V) = R(V) \\; \\times \\; E(B-V)

            THRU = 10^{-0.4 \\; A(V)}

        Parameters
        ----------
        ebv : float or `~astropy.units.quantity.Quantity`
            :math:`E(B-V)` value in magnitude.

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` is used.

        Returns
        -------
        extcurve : `ExtinctionCurve`
            Empirical extinction curve.

        Raises
        ------
        synphot.exceptions.SynphotError
            Invalid input.

        """
        if isinstance(ebv, u.Quantity) and ebv.unit.decompose() == u.mag:
            ebv = ebv.value
        elif not isinstance(ebv, numbers.Real):
            raise exceptions.SynphotError('E(B-V)={0} is invalid.'.format(ebv))

        x = self._validate_wavelengths(wavelengths)
        header = {'E(B-V)': ebv}

        # Duck-typing dust-extinction package API.
        if HAS_DUST_EXTINCTION and hasattr(self.model, 'extinguish'):
            y = self.model.extinguish(x, Ebv=ebv)
            header['ReddeningLaw'] = '{!r}'.format(self.model)
        else:
            y = 10 ** (-0.4 * self(x).value * ebv)
            header['ReddeningLaw'] = self.meta.get('expr', 'unknown')

        return ExtinctionCurve(ExtinctionModel1D, points=x, lookup_table=y,
                               meta={'header': header})

    def to_fits(self, filename, wavelengths=None, **kwargs):
        """Write the reddening law to a FITS file.

        :math:`R(V)` column is automatically named 'Av/E(B-V)'.

        Parameters
        ----------
        filename : str
            Output filename.

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` is used.

        kwargs : dict
            Keywords accepted by :func:`~synphot.specio.write_fits_spec`.

        """
        w, y = self._get_arrays(wavelengths)

        kwargs['flux_col'] = 'Av/E(B-V)'
        kwargs['flux_unit'] = self._internal_flux_unit

        # No need to trim/pad zeroes, unless user chooses to do so.
        if 'pad_zero_ends' not in kwargs:
            kwargs['pad_zero_ends'] = False
        if 'trim_zero' not in kwargs:
            kwargs['trim_zero'] = False

        # There are some standard keywords that should be added
        # to the extension header.
        bkeys = {'tdisp1': 'G15.7', 'tdisp2': 'G15.7'}

        if 'expr' in self.meta:
            bkeys['expr'] = (self.meta['expr'], 'synphot expression')

        if 'ext_header' in kwargs:
            kwargs['ext_header'].update(bkeys)
        else:
            kwargs['ext_header'] = bkeys

        specio.write_fits_spec(filename, w, y, **kwargs)

    @classmethod
    def from_file(cls, filename, **kwargs):
        """Create a reddening law from file.

        If filename has 'fits' or 'fit' suffix, it is read as FITS.
        Otherwise, it is read as ASCII.

        Parameters
        ----------
        filename : str
            Reddening law filename.

        kwargs : dict
            Keywords acceptable by
            :func:`~synphot.specio.read_fits_spec` (if FITS) or
            :func:`~synphot.specio.read_ascii_spec` (if ASCII).

        Returns
        -------
        redlaw : `ReddeningLaw`
            Empirical reddening law.

        """
        if 'flux_unit' not in kwargs:
            kwargs['flux_unit'] = cls._internal_flux_unit

        if ((filename.endswith('fits') or filename.endswith('fit')) and
                'flux_col' not in kwargs):
            kwargs['flux_col'] = 'Av/E(B-V)'

        header, wavelengths, rvs = specio.read_spec(filename, **kwargs)

        return cls(Empirical1D, points=wavelengths, lookup_table=rvs,
                   meta={'header': header})

    @classmethod
    def from_extinction_model(cls, modelname, **kwargs):
        """Load :ref:`pre-defined extinction model <synphot_reddening>`.

        Parameters
        ----------
        modelname : str
            Extinction model name. Choose from 'lmc30dor', 'lmcavg', 'mwavg',
            'mwdense', 'mwrv21', 'mwrv40', 'smcbar', or 'xgalsb'.

        kwargs : dict
            Keywords acceptable by :func:`~synphot.specio.read_remote_spec`.

        Returns
        -------
        redlaw : `ReddeningLaw`
            Empirical reddening law.

        Raises
        ------
        synphot.exceptions.SynphotError
            Invalid extinction model name.

        """
        modelname = modelname.lower()

        # Select filename based on model name
        if modelname == 'lmc30dor':
            cfgitem = Conf.lmc30dor_file
        elif modelname == 'lmcavg':
            cfgitem = Conf.lmcavg_file
        elif modelname == 'mwavg':
            cfgitem = Conf.mwavg_file
        elif modelname == 'mwdense':
            cfgitem = Conf.mwdense_file
        elif modelname == 'mwrv21':
            cfgitem = Conf.mwrv21_file
        elif modelname == 'mwrv40':
            cfgitem = Conf.mwrv40_file
        elif modelname == 'smcbar':
            cfgitem = Conf.smcbar_file
        elif modelname == 'xgalsb':
            cfgitem = Conf.xgal_file
        else:
            raise exceptions.SynphotError(
                'Extinction model {0} is invalid.'.format(modelname))

        filename = cfgitem()

        if 'flux_unit' not in kwargs:
            kwargs['flux_unit'] = cls._internal_flux_unit

        if ((filename.endswith('fits') or filename.endswith('fit')) and
                'flux_col' not in kwargs):
            kwargs['flux_col'] = 'Av/E(B-V)'

        header, wavelengths, rvs = specio.read_remote_spec(filename, **kwargs)
        header['filename'] = filename
        header['descrip'] = cfgitem.description
        meta = {'header': header, 'expr': modelname}

        return cls(Empirical1D, points=wavelengths, lookup_table=rvs,
                   meta=meta)


class ExtinctionCurve(BaseUnitlessSpectrum):
    """Class to handle extinction curve.

    Parameters
    ----------
    modelclass, kwargs
        See `~synphot.spectrum.BaseSpectrum`.

    """
    pass


# TODO: Find a better way to handle so many magic numbers.
# See https://github.com/spacetelescope/synphot_refactor/issues/77
def etau_madau(wave, z, **kwargs):
    """Madau 1995 extinction for a galaxy at given redshift.
    This is the Lyman-alpha prescription from the photo-z code BPZ.

    The Lyman-alpha forest approximately has an effective
    "throughput" which is a function of redshift and
    rest-frame wavelength.
    One would multiply the SEDs by this factor before
    passing it through an instrument filter.

    This approximation is from Footnote 3 of
    :ref:`Madau et al. (1995) <synphot-ref-madau1995>`.
    This is claimed accurate to 5%.
    The scatter in this factor (due to different lines of sight)
    is huge, as shown in Madau's Fig. 3 (top panel);
    The figure's bottom panel shows a redshifted version of the
    "exact" prescription.

    Parameters
    ----------
    wave : array-like or `~astropy.units.quantity.Quantity`
        Redshifted wavelength values.
        Non-redshifted wavelength is ``wave / (1 + z)``.

    z : number
        Redshift.

    kwargs : dict
        Equivalencies for unit conversion, see
        :func:`~synphot.units.validate_quantity`.

    Returns
    -------
    extcurve : `ExtinctionCurve`
        Extinction curve to apply to the redshifted spectrum.

    """
    if not isinstance(z, numbers.Real):
        raise exceptions.SynphotError(
            'Redshift must be a real scalar number.')

    if np.isscalar(wave) or len(wave) <= 1:
        raise exceptions.SynphotError('Wavelength has too few data points')

    wave = units.validate_quantity(wave, u.AA, **kwargs).value

    ll = 912.0
    c = np.array([3.6e-3, 1.7e-3, 1.2e-3, 9.3e-4])
    el = np.array([1216, 1026, 973, 950], dtype=float)  # noqa
    tau = np.zeros_like(wave, dtype=float)
    xe = 1.0 + z

    # Lyman series
    for i in range(len(el)):
        tau = np.where(wave <= el[i] * xe,
                       tau + c[i] * (wave / el[i]) ** 3.46,
                       tau)

    # Photoelectric absorption
    xc = wave / ll
    xc3 = xc ** 3
    tau = np.where(wave <= ll * xe,
                   (tau + 0.25 * xc3 * (xe ** 0.46 - xc ** 0.46) +
                    9.4 * xc ** 1.5 * (xe ** 0.18 - xc ** 0.18) -
                    0.7 * xc3 * (xc ** (-1.32) - xe ** (-1.32)) -
                    0.023 * (xe ** 1.68 - xc ** 1.68)),
                   tau)

    thru = np.where(tau > 700., 0., np.exp(-tau))
    meta = {'descrip': 'Madau 1995 extinction for z={0}'.format(z)}
    return ExtinctionCurve(ExtinctionModel1D, points=wave, lookup_table=thru,
                           meta=meta)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities related to wavelength bin calculations."""

# THIRD-PARTY
import numpy as np

# ASTROPY
from astropy import units as u

# LOCAL
from synphot import exceptions

__all__ = ['calculate_bin_edges', 'calculate_bin_widths',
           'calculate_bin_centers', 'wave_range', 'pixel_range']


def _slow_calcbinflux(len_binwave, i_beg, i_end, avflux, deltaw):
    """Python implementation of ``calcbinflux``.

    This is only used if ``synphot.synphot_utils`` C-extension
    import fails.

    See ``include/synphot_utils.h``.

    """
    binflux = np.empty(shape=(len_binwave, ), dtype=np.float64)
    intwave = np.empty(shape=(len_binwave, ), dtype=np.float64)

    # Note that, like all Python striding, the range over which
    # we integrate is [first:last).
    for i in range(len(i_beg)):
        first = i_beg[i]
        last = i_end[i]
        cur_dw = deltaw[first:last]
        intwave[i] = cur_dw.sum()
        binflux[i] = np.sum(avflux[first:last] * cur_dw) / intwave[i]

    return binflux, intwave


# Try to import the C version of calcbinflux, otherwise fall back
# to the Python implementation above.
try:
    from synphot import synphot_utils
except ImportError:
    calcbinflux = _slow_calcbinflux
else:
    calcbinflux = synphot_utils.calcbinflux


def calculate_bin_edges(centers):
    """Calculate the edges of wavelength bins given the centers.

    The algorithm calculates bin edges as the midpoints between bin centers
    and treats the first and last bins as symmetric about their centers.

    Parameters
    ----------
    centers : array-like or `~astropy.units.quantity.Quantity`
        Sequence of bin centers. Must be 1D and have at least two values.
        If not a Quantity, assumed to be in Angstrom.

    Returns
    -------
    edges : `~astropy.units.quantity.Quantity`
        Array of bin edges. Will be 1D, have one more value
        than ``centers``, and also the same unit.

    Raises
    ------
    synphot.exceptions.SynphotError
        Invalid input.

    """
    if not isinstance(centers, u.Quantity):
        centers = centers * u.AA

    if centers.ndim != 1:
        raise exceptions.SynphotError('Bin centers must be 1D array.')

    if centers.size < 2:
        raise exceptions.SynphotError(
            'Bin centers must have at least two values.')

    edges = np.empty(centers.size + 1, dtype=np.float64)
    edges[1:-1] = (centers.value[1:] + centers.value[:-1]) * 0.5

    # Compute the first and last by making them symmetric
    edges[0] = 2.0 * centers.value[0] - edges[1]
    edges[-1] = 2.0 * centers.value[-1] - edges[-2]

    return edges * centers.unit


def calculate_bin_widths(edges):
    """Calculate the widths of wavelengths bins given their edges.

    Parameters
    ----------
    edges : array-like or `~astropy.units.quantity.Quantity`
        Sequence of bin edges. Must be 1D and have at least two values.
        If not a Quantity, assumed to be in Angstrom.

    Returns
    -------
    widths : `~astropy.units.quantity.Quantity`
        Array of bin widths. Will be 1D, have one less value
        than ``edges``, and also the same unit.

    Raises
    ------
    synphot.exceptions.SynphotError
        Invalid input.

    """
    if not isinstance(edges, u.Quantity):
        edges = edges * u.AA

    if edges.ndim != 1:
        raise exceptions.SynphotError('Bin edges must be 1D array.')

    if edges.size < 2:
        raise exceptions.SynphotError(
            'Bin edges must have at least two values.')

    return np.abs(edges[1:] - edges[:-1])


def calculate_bin_centers(edges):
    """Calculate the centers of wavelengths bins given their edges.

    Parameters
    ----------
    edges : array-like or `~astropy.units.quantity.Quantity`
        Sequence of bin edges. Must be 1D and have at least two values.
        If not a Quantity, assumed to be in Angstrom.

    Returns
    -------
    centers : `~astropy.units.quantity.Quantity`
        Array of bin centers. Will be 1D, have one less value
        than ``edges``, and also the same unit.

    Raises
    ------
    synphot.exceptions.SynphotError
        Invalid input.

    """
    if not isinstance(edges, u.Quantity):
        edges = edges * u.AA

    if edges.ndim != 1:
        raise exceptions.SynphotError('Bin edges must be 1D array.')

    if edges.size < 2:
        raise exceptions.SynphotError(
            'Bin edges must have at least two values.')

    centers = np.empty(edges.size - 1, dtype=np.float64)
    centers[0] = edges.value[:2].mean()

    for i in range(1, centers.size):
        centers[i] = 2.0 * edges.value[i] - centers[i - 1]

    return centers * edges.unit


def wave_range(bins, cenwave, npix, mode='round'):
    """Calculate the wavelength range covered by the given number of pixels
    centered on the given central wavelength of the given bins.

    Parameters
    ----------
    bins : array-like
        Wavelengths at bin centers, each centered on a pixel.
        Must be 1D array.

    cenwave : float
        Desired central wavelength, in the same unit as ``bins``.

    npix : int
        Desired number of pixels, centered on ``cenwave``.

    mode : {'round', 'min', 'max', 'none'}
        Determines how the pixels at the edges of the wavelength range
        are handled. All the options, except 'none', will return
        wavelength range edges that correspond to pixel edges:

            * 'round' - Wavelength range edges are the pixel edges
              and the range spans exactly ``npix`` pixels. An edge
              that falls in the center of a bin is rounded to the
              nearest pixel edge. This is the default.

            * 'min' - Wavelength range is shrunk such that it includes
              an integer number of pixels and its edges fall on pixel
              edges. It may not span exactly ``npix`` pixels.

            * 'max' - Wavelength range is expanded such that it
              includes an integer number of pixels and its edges fall
              on pixel edges. It may not span exactly ``npix`` pixels.

            * 'none' - Exact wavelength range is returned. The edges
              may not correspond to pixel edges, but it covers exactly
              ``npix`` pixels.

    Returns
    -------
    wave1, wave2 : float
        Lower and upper limits of the wavelength range.

    Raises
    ------
    synphot.exceptions.OverlapError
        Given central wavelength is not within the given bins
        or the wavelength range would exceed the bin limits.

    synphot.exceptions.SynphotError
        Invalid inputs or calculation failed.

    """
    mode = mode.lower()

    if mode not in ('round', 'min', 'max', 'none'):
        raise exceptions.SynphotError(
            'mode={0} is invalid, must be "round", "min", "max", '
            'or "none".'.format(mode))

    if not isinstance(npix, int):
        raise exceptions.SynphotError('npix={0} is invalid.'.format(npix))

    # Bin values must be in ascending order.
    if bins[0] > bins[-1]:
        bins = bins[::-1]

    # Central wavelength must be within given bins.
    if cenwave < bins[0] or cenwave > bins[-1]:
        raise exceptions.OverlapError(
            'cenwave={0} is not within binset (min={1}, max={2}).'.format(
                cenwave, bins[0], bins[-1]))

    # Find the index the central wavelength among bins
    diff = cenwave - bins
    ind = np.argmin(np.abs(diff))

    # Calculate fractional index
    frac_ind = float(ind)
    if diff[ind] < 0:
        frac_ind += diff[ind] / (bins[ind] - bins[ind - 1])
    elif diff[ind] > 0:
        frac_ind += diff[ind] / (bins[ind + 1] - bins[ind])

    # Calculate fractional indices of the edges
    half_npix = npix / 2.0
    frac_ind1 = frac_ind - half_npix
    frac_ind2 = frac_ind + half_npix

    # Calculated edges must not exceed bin edges
    if frac_ind1 < -0.5:
        raise exceptions.OverlapError(
            'Lower limit of wavelength range is out of bounds.')
    if frac_ind2 > (bins.size - 0.5):
        raise exceptions.OverlapError(
            'Upper limit of wavelength range is out of bounds.')

    frac1, int1 = np.modf(frac_ind1)
    frac2, int2 = np.modf(frac_ind2)
    int1 = int(int1)
    int2 = int(int2)

    if mode == 'round':
        # Lower end of wavelength range
        if frac1 >= 0:
            # end is somewhere greater than binset[0] so we can just
            # interpolate between two neighboring values going with upper edge
            wave1 = bins[int1:int1 + 2].mean()
        else:
            # end is below the lowest binset value, but not by enough to
            # trigger an exception
            wave1 = bins[0] - (bins[0:2].mean() - bins[0])

        # Upper end of wavelength range
        if int2 < bins.shape[0] - 1:
            # end is somewhere below binset[-1] so we can just interpolate
            # between two neighboring values, going with the upper edge.
            wave2 = bins[int2:int2 + 2].mean()
        else:
            # end is above highest binset value but not by enough to
            # trigger an exception
            wave2 = bins[-1] + (bins[-1] - bins[-2:].mean())

    elif mode == 'min':
        # Lower end of wavelength range
        if frac1 <= 0.5 and int1 < bins.shape[0] - 1:
            # not at the lowest possible edge and pixel i included
            wave1 = bins[int1:int1 + 2].mean()
        elif frac1 > 0.5 and int1 < bins.shape[0] - 2:
            # not at the lowest possible edge and pixel i not included
            wave1 = bins[int1 + 1:int1 + 3].mean()
        elif frac1 == -0.5:
            # at the lowest possible edge
            wave1 = bins[0] - (bins[0:2].mean() - bins[0])
        else:  # pragma: no cover
            raise exceptions.SynphotError(
                'mode={0} gets unexpected frac1={1}, int1={2}'.format(
                    mode, frac1, int1))

        # Upper end of wavelength range
        if frac2 >= 0.5 and int2 < bins.shape[0] - 1:
            # not out at the end and pixel i included
            wave2 = bins[int2:int2 + 2].mean()
        elif frac2 < 0.5 and int2 < bins.shape[0]:
            # not out at end and pixel i not included
            wave2 = bins[int2 - 1:int2 + 1].mean()
        elif frac2 == 0.5 and int2 == bins.shape[0] - 1:
            # at the very end
            wave2 = bins[-1] + (bins[-1] - bins[-2:].mean())
        else:  # pragma: no cover
            raise exceptions.SynphotError(
                'mode={0} gets unexpected frac2={1}, int2={2}'.format(
                    mode, frac2, int2))

    elif mode == 'max':
        # Lower end of wavelength range
        if frac1 < 0.5 and int1 < bins.shape[0]:
            # not at the lowest possible edge and pixel i included
            wave1 = bins[int1 - 1:int1 + 1].mean()
        elif frac1 >= 0.5 and int1 < bins.shape[0] - 1:
            # not at the lowest possible edge and pixel i not included
            wave1 = bins[int1:int1 + 2].mean()
        elif frac1 == -0.5:
            # at the lowest possible edge
            wave1 = bins[0] - (bins[0:2].mean() - bins[0])
        else:  # pragma: no cover
            raise exceptions.SynphotError(
                'mode={0} gets unexpected frac1={1}, int1={2}'.format(
                    mode, frac1, int1))

        # Upper end of wavelength range
        if frac2 > 0.5 and int2 < bins.shape[0] - 2:
            # not out at the end and pixel i included
            wave2 = bins[int2 + 1:int2 + 3].mean()
        elif frac2 <= 0.5 and int2 < bins.shape[0] - 1:
            # not out at end and pixel i not included
            wave2 = bins[int2:int2 + 2].mean()
        elif frac2 == 0.5 and int2 == bins.shape[0] - 1:
            # at the very end
            wave2 = bins[-1] + (bins[-1] - bins[-2:].mean())
        else:  # pragma: no cover
            raise exceptions.SynphotError(
                'mode={0} gets unexpected frac2={1}, int2={2}'.format(
                    mode, frac2, int2))

    else:  # mode == 'none'
        wave1 = bins[int1] + frac1 * (bins[int1 + 1] - bins[int1])
        wave2 = bins[int2] + frac2 * (bins[int2 + 1] - bins[int2])

    return wave1, wave2


def pixel_range(bins, waverange, mode='round'):
    """Calculate the number of pixels within the given wavelength range
    and the given bins.

    Parameters
    ----------
    bins : array-like
        Wavelengths at bin centers, each centered on a pixel.
        Must be 1D array.

    waverange : tuple of float
        Lower and upper limits of the desired wavelength range,
        in the same unit as ``bins``.

    mode : {'round', 'min', 'max', 'none'}
        Determines how the pixels at the edges of the wavelength range
        are handled. All the options, except 'none', will return
        an integer number of pixels:

            * 'round' - Wavelength range edges that fall in the middle
              of a pixel are counted if more than half of the pixel is
              within the given wavelength range. Edges that fall in
              the center of a pixel are rounded to the nearest pixel
              edge. This is the default.

            * 'min' - Only pixels wholly within the given wavelength
              range are counted.

            * 'max' - Pixels that are within the given wavelength range
              by any margin are counted.

            * 'none' - The exact number of encompassed pixels,
              including fractional pixels, is returned.

    Returns
    -------
    npix : number
        Number of pixels.

    Raises
    ------
    synphot.exceptions.OverlapError
        Given wavelength range exceeds the bounds of given bins.

    synphot.exceptions.SynphotError
        Invalid mode.

    """
    mode = mode.lower()

    if mode not in ('round', 'min', 'max', 'none'):
        raise exceptions.SynphotError(
            'mode={0} is invalid, must be "round", "min", "max", '
            'or "none".'.format(mode))

    if waverange[0] < waverange[-1]:
        wave1 = waverange[0]
        wave2 = waverange[-1]
    else:
        wave1 = waverange[-1]
        wave2 = waverange[0]

    # Bin values must be in ascending order.
    if bins[0] > bins[-1]:
        bins = bins[::-1]

    # Wavelength range must be within bins
    minwave = bins[0] - (bins[0:2].mean() - bins[0])
    maxwave = bins[-1] + (bins[-1] - bins[-2:].mean())
    if wave1 < minwave or wave2 > maxwave:
        raise exceptions.OverlapError(
            'Wavelength range ({0}, {1}) is out of bounds of bins '
            '(min={2}, max={3}).'.format(wave1, wave2, minwave, maxwave))

    if wave1 == wave2:
        return 0

    if mode == 'round':
        ind1 = bins.searchsorted(wave1, side='right')
        ind2 = bins.searchsorted(wave2, side='right')
    else:
        ind1 = bins.searchsorted(wave1, side='left')
        ind2 = bins.searchsorted(wave2, side='left')

    if mode == 'round':
        npix = ind2 - ind1

    elif mode == 'min':
        # for ind1, figure out if pixel ind1 is wholly included or not.
        # do this by figuring out where wave1 is between ind1 and ind1-1.
        frac = (bins[ind1] - wave1) / (bins[ind1] - bins[ind1 - 1])
        if frac < 0.5:
            # ind1 is only partially included
            ind1 += 1

        # similar but reversed procedure for ind2
        frac = (wave2 - bins[ind2 - 1]) / (bins[ind2] - bins[ind2 - 1])
        if frac < 0.5:
            # ind2 is only partially included
            ind2 -= 1

        npix = ind2 - ind1

    elif mode == 'max':
        # for ind1, figure out if pixel ind1-1 is partially included or not.
        # do this by figuring out where wave1 is between ind1 and ind1-1.
        frac = (wave1 - bins[ind1 - 1]) / (bins[ind1] - bins[ind1 - 1])
        if frac < 0.5:
            # ind1 is partially included
            ind1 -= 1

        # similar but reversed procedure for ind2
        frac = (bins[ind2] - wave2) / (bins[ind2] - bins[ind2 - 1])
        if frac < 0.5:
            # ind2 is partially included
            ind2 += 1

        npix = ind2 - ind1

    else:  # mode == 'none'
        # calculate fractional indices
        frac1 = ind1 - (bins[ind1] - wave1) / (bins[ind1] - bins[ind1 - 1])
        frac2 = ind2 - (bins[ind2] - wave2) / (bins[ind2] - bins[ind2 - 1])
        npix = frac2 - frac1

    return npix



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test reddening.py module."""

# STDLIB
import os
import shutil
import tempfile

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.io import fits
from astropy.tests.helper import assert_quantity_allclose
from astropy.units import add_enabled_equivalencies
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning

# LOCAL
from synphot import exceptions, units
from synphot.compat import HAS_DUST_EXTINCTION  # noqa
from synphot.models import ConstFlux1D, Empirical1D
from synphot.observation import Observation
from synphot.reddening import ReddeningLaw, etau_madau
from synphot.spectrum import SourceSpectrum, SpectralElement


class TestExtinction:
    """Test ReddeningLaw and ExtinctionCurve classes (most methods)."""
    def setup_class(self):
        rfile = get_pkg_data_filename(
            os.path.join('data', 'milkyway_diffuse_001.fits'),
            package='synphot.tests')
        self.redlaw = ReddeningLaw.from_file(rfile)
        self.extcurve = self.redlaw.extinction_curve(0.3 * u.mag)

    def test_invalid_ebv(self):
        with pytest.raises(exceptions.SynphotError):
            self.redlaw.extinction_curve(1 * units.FLAM)

    def test_redlaw_call(self):
        w = self.redlaw.waveset[48:53]
        with add_enabled_equivalencies(u.spectral()):
            assert_quantity_allclose(
                w, [5.41599989, 5.3204999, 5.2249999, 5.12949991,
                    5.03399992] * (u.micron ** -1))
        assert_quantity_allclose(
            self.redlaw(w),
            [7.90572977, 8.01734924, 8.17892265, 8.40150452, 8.69231796])

    def test_extcurve_call(self):
        # Cannot use waveset.
        # See https://github.com/spacetelescope/synphot_refactor/issues/129
        w = np.squeeze(self.extcurve.model.points) * u.AA

        y = self.extcurve(w)
        np.testing.assert_array_equal(w, self.redlaw.waveset)
        np.testing.assert_allclose(y, 10 ** (-0.4 * self.redlaw(w) * 0.3))

    @pytest.mark.skipif('not HAS_DUST_EXTINCTION')
    def test_extcurve_call_2(self):
        from dust_extinction.parameter_averages import CCM89

        Av = 2
        wav = np.arange(0.1, 3, 0.001) * u.micron
        ext = CCM89(Rv=3.1)
        ans = ext.extinguish(wav, Av)

        ebv = Av / ext.Rv
        redlaw = ReddeningLaw(ext)
        extcurve = redlaw.extinction_curve(ebv, wavelengths=wav)

        assert_quantity_allclose(extcurve(wav), ans)
        assert extcurve.meta['header']['ReddeningLaw'] == '<CCM89(Rv=3.1)>'

    def test_mul_spec(self):
        """Apply extinction curve in inverse micron to flat spectrum in
        Angstrom.

        """
        sp = SourceSpectrum(ConstFlux1D, amplitude=1)
        sp2 = self.extcurve * sp
        w = 5.03399992 * (u.micron ** -1)
        ans = self.extcurve(w).value
        np.testing.assert_allclose(sp2(w).value, ans, rtol=1e-6)

    def test_qso_countrate(self):
        """Ensure extinction curve waveset is not propagated to spectrum.
        https://github.com/spacetelescope/synphot_refactor/issues/129
        """
        bp = SpectralElement.from_file(get_pkg_data_filename(
            os.path.join('data', 'hst_acs_hrc_f850lp.fits'),
            package='synphot.tests'))
        qso = SourceSpectrum.from_file(get_pkg_data_filename(
            os.path.join('data', 'qso_template_001.dat'),
            package='synphot.tests'))
        extcurve = self.redlaw.extinction_curve(1.0 * u.mag)
        spext = qso * extcurve
        with pytest.warns(AstropyUserWarning):
            sp = spext.normalize(25 * u.STmag, bp)
            obs = Observation(sp, bp, force='taper')
        area = 45238.93416  # HST cm^2
        c = obs.countrate(area)
        ans = 1.104404103799421e-07  # From similar setup in Astrolib PySynphot
        np.testing.assert_allclose(c.value, ans, rtol=1e-3)  # 0.1% agreement


# See https://github.com/spacetelescope/synphot_refactor/issues/77
@pytest.mark.parametrize(
    ('z', 'ans'),
    ([0, [0.99514224, 0.99572959, 0.99630696, 0.99640647, 1]],
     [2, [0.80417561, 0.82569455, 0.84739614, 0.85119226, 1]],
     [4, [0.27908754, 0.32576126, 0.37920904, 0.38926572, 1]],
     [8, [5.80763088e-05, 1.89352199e-04, 6.04679639e-04, 7.38588957e-04, 1]]))
def test_etau_madau(z, ans):
    """Test Madau 1995 extinction curve."""
    w_rest = np.array([950, 973, 1026, 1216, 1300])
    w_z = w_rest * (1 + z)
    extcurve = etau_madau(w_z, z)
    np.testing.assert_allclose(extcurve(w_z), ans)


def test_etau_madau_exceptions():
    # Invalid z
    with pytest.raises(exceptions.SynphotError):
        etau_madau([500, 1000], [1, 2])

    # Too few wave
    with pytest.raises(exceptions.SynphotError):
        etau_madau(500, 0)
    with pytest.raises(exceptions.SynphotError):
        etau_madau([500], 0)


@pytest.mark.remote_data
@pytest.mark.parametrize(
    'modelname',
    ['lmc30dor', 'lmcavg', 'mwavg', 'mwdense', 'mwrv21', 'mwrv40',
     'smcbar', 'xgalsb'])
def test_redlaw_from_model(modelname):
    """Test ReddeningLaw from remote file.

    .. note:: No check on data quality as it is dependent on data file.

    """
    redlaw = ReddeningLaw.from_extinction_model(modelname)
    assert modelname in redlaw.meta['expr']
    assert 'filename' in redlaw.meta['header']
    assert 'descrip' in redlaw.meta['header']


def test_redlaw_from_model_exception():
    with pytest.raises(exceptions.SynphotError):
        ReddeningLaw.from_extinction_model('foo')


class TestWriteReddeningLaw:
    """Test ReddeningLaw ``to_fits()`` method."""
    def setup_class(self):
        self.outdir = tempfile.mkdtemp()
        self.x = np.linspace(1000, 5000, 5)
        self.y = np.linspace(1, 5, 5) * 0.1
        self.redlaw = ReddeningLaw(
            Empirical1D, points=self.x, lookup_table=self.y)

    @pytest.mark.parametrize('ext_hdr', [None, {'foo': 'foo'}])
    def test_write(self, ext_hdr):
        outfile = os.path.join(self.outdir, 'outredlaw.fits')

        if ext_hdr is None:
            self.redlaw.to_fits(outfile, overwrite=True)
        else:
            self.redlaw.to_fits(outfile, overwrite=True, ext_header=ext_hdr)

        # Read it back in and check
        redlaw2 = ReddeningLaw.from_file(outfile)
        np.testing.assert_allclose(redlaw2.waveset.value, self.x)
        np.testing.assert_allclose(redlaw2(self.x).value, self.y)

        if ext_hdr is not None:
            hdr = fits.getheader(outfile, 1)
            assert 'foo' in hdr

    def teardown_class(self):
        shutil.rmtree(self.outdir)
